#include "ppc.h"

#include "m33.h"

#include "framebuffer.h"

PPC::PPC(float hfov, int _w, int _h): a(1.0f, 0.0f, 0.0f), b(0.0f, -1.0f, 0.0f),
	C(0.0f, 0.0f, 0.0f), w(_w), h(_h) {

	float hfovr = hfov * 3.14159f / 180.0f;
	float f = -(float)w / (2.0f *tanf(hfovr / 2.0f));
	c = V3(-(float)w / 2.0f, (float)h / 2.0f, f);

}


int PPC::Project(V3 p, V3& pp) {

	M33 M; 
	M.SetColumn(0, a);
	M.SetColumn(1, b);
	M.SetColumn(2, c);
	pp[0] = FLT_MAX;
	V3 q = M.Inverted()*(p - C);
	if (q[2] < 0.0f)
		return 0;
	pp[2] = 1.0f / q[2];
	pp[0] = q[0] / q[2];
	pp[1] = q[1] / q[2];
	return 1;

}

V3 PPC::GetVD() {

	return (a ^ b).UnitVector();

}

float PPC::GetFocalLength() {

	return GetVD()*c;

}

void PPC::ChangeFocalLength(float scf) {

	c = c + GetVD()*(GetFocalLength()*(scf - 1.0f));

}


void PPC::PositionAndOrient(V3 C1, V3 L1, V3 vpv) {

	V3 vd1 = (L1 - C1).UnitVector();
	V3 a1 = (vd1 ^ vpv).UnitVector() * a.Length();
	V3 b1 = (vd1 ^ a1).UnitVector() * b.Length();
	V3 c1 = vd1 * GetFocalLength() - b1 * ((float)h / 2.0f) - a1 * ((float)w / 2.0f);

	C = C1;
	a = a1;
	b = b1;
	c = c1;

}


void PPC::Pan(float angled) {


	V3 dv = (b * -1.0f).UnitVector();
	a = a.RotateThisVectorAboutDirection(dv, angled);
	c = c.RotateThisVectorAboutDirection(dv, angled);

}



void PPC::Visualize(PPC *ppc3, FrameBuffer *fb3, float vf) {

	V3 colv(0.0f, 0.0f, 0.0f);
	fb3->Draw3DPoint(C, colv, ppc3, 7);

	float scf = vf / GetFocalLength();
	fb3->Draw3DSegment(C, colv, C + c * scf, colv, ppc3);
	fb3->Draw3DSegment(C + c * scf, colv, C + (c + a * (float)w)*scf, colv, ppc3);
	fb3->Draw3DSegment(C + (c + a * (float)w)*scf, colv,
		C + (c + a * (float)w + b * (float)h)*scf, colv, ppc3);
	fb3->Draw3DSegment(
		C + (c + a * (float)w + b * (float)h)*scf, colv,
		C + (c +  b * (float)h)*scf, colv,
		ppc3);
	fb3->Draw3DSegment(
		C + (c + b * (float)h)*scf, colv,
		C+c*scf, colv,
		ppc3);

}


V3 PPC::UnprojectPixel(float uf, float vf, float currf) {

	return C + (a*uf + b * vf + c)*currf * (1.0f /  GetFocalLength());

}


V3 PPC::Unproject(V3 pP) {

	return C + (a*pP[0] + b * pP[1] + c) * (1.0f / pP[2]);

}



void PPC::SetIntrinsicsHW() {

	glViewport(0, 0, w, h);
	float zNear = 1.0f;
	float zFar = 1000.0f;
	float scf = zNear / GetFocalLength();
	float left = -a.Length()*(float)w / 2.0f*scf;
	float right = a.Length()*(float)w / 2.0f*scf;
	float top = b.Length()*(float)h / 2.0f*scf;
	float bottom = -b.Length()*(float)h / 2.0f*scf;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(left, right, bottom, top, zNear, zFar);
	glMatrixMode(GL_MODELVIEW);

}
void PPC::SetExtrinsicsHW() {

	V3 L = C + GetVD()*100.0f;
	glLoadIdentity();
	gluLookAt(C[0], C[1], C[2], L[0], L[1], L[2], -b[0], -b[1], -b[2]);

}
