#include "scene.h"
#include "v3.h"
#include "m33.h"
#include "ppc.h"
#include "tm.h"
#include <stdlib.h>

Scene *scene;

using namespace std;

#include <fstream>

#include <iostream>

Scene::Scene() {

	L = V3(0.0f, 0.0f, 0.0f);

	tms = 0;
	tmsN = 0;

	gui = new GUI();
	gui->show();

	int u0 = 20;
	int v0 = 20;
	int w = 640;
	int h = 480;

	fb = new FrameBuffer(u0, v0, w, h);
	fb->label("SW Framebuffer");
	fb->show();

	hwfb = new FrameBuffer(u0 + fb->w + 30, v0, w, h);
	hwfb->label("HW Framebuffer");
	hwfb->ishw = true;
	hwfb->show();

	fb3 = new FrameBuffer(u0 + fb->w + 30, v0, w, h);
	fb3->label("Third person");
//	fb3->show();

	float hfov = 55.0f;
	ppc = new PPC(hfov, fb->w, fb->h);
	ppc3 = new PPC(30.0f, fb3->w, fb3->h);

	gui->uiw->position(u0, v0 + fb->h + 60);

	tmsN = 1;
	tms = new TM[tmsN];
	tms[0].LoadBin("geometry/teapot1K.bin");
	V3 tmC = ppc->C + ppc->GetVD()*100.0f;
	float tmSize = 100.0f;
	tms[0].PositionAndSize(tmC, tmSize);


	ppc3->C = ppc3->C + V3(330.0f, 150.0f, 300.0f);
	ppc3->PositionAndOrient(ppc3->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));

	Render();

}

void Scene::RenderHW() {

	// clear the frame buffers
	glEnable(GL_DEPTH_TEST);

	glClearColor(0.0f, 0.0f, 0.5f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// set camera instrinsics
	ppc->SetIntrinsicsHW();
	// set camera extrinsics
	ppc->SetExtrinsicsHW();

	// render geometry
	for (int tmi = 0; tmi < tmsN; tmi++) {
		tms[tmi].RenderHW();
	}

}

void Scene::Render() {

	if (fb) {
		Render(ppc, fb);
	}

	float currf = 40.0f;
	if (fb3) {
		fb3->SetBGR(0xFFFFFFFF);
		fb3->ClearZB(0.0f);
		Render(ppc3, fb3);
		fb3->Draw3DPoint(L, V3(1.0f, 1.0f, 0.3f), ppc3, 7);
//		ppc->Visualize(ppc3, fb3, currf);
//		fb->Visualize(ppc, currf, ppc3, fb3);
//		fb->Visualize3D(ppc, currf, ppc3, fb3);
	}

}

void Scene::Render(PPC *currppc, FrameBuffer *currfb) {

	currfb->SetBGR(0xFFFFFFFF);
	currfb->ClearZB(0.0f);
	for (int tmi = 0; tmi < tmsN; tmi++) {
		tms[tmi].RenderFilled(currppc, currfb);
//		tms[tmi].RenderWireframe(currppc, currfb);
//		tms[tmi].RenderPoints(currppc, currfb);
	}
	currfb->redraw();

}

void Scene::DBG() {

	{
		for (int i = 0; i < 100; i++) {
			ppc->C = ppc->C + ppc->a.UnitVector()*0.1f;
			Render(ppc, fb);
			hwfb->redraw();
			Fl::check();
		}
		return;

	}

	{

		fb->SetBGR(0xFFFFFFFF);
		fb->ClearZB(0.0f);
		tms[0].RayTrace(ppc, fb);
		fb->redraw();
		return;
	}

	{

		L = ppc->C;
		tms[0].Light(V3(1.0f, 0.0f, 0.0f), L);
		for (int i = 0; i < 360; i++) {
			kse = 1000.0f * ((float) i) / 359.0f;
			Render();
			Fl::check();
		}
		cerr << "INFO: done" << endl;
		return;

	}

	{
		V3 tc = tms[0].GetCenter();
		for (int i = 0; i < 360; i++) {
			L = tc + V3(40.0f, 0.0f, 0.0f);
			L = L.RotateThisPointAboutArbitraryAxis(tc, V3(0.0f, 1.0f, 0.0f), (float)(i*2));
			tms[0].Light(V3(1.0f, 0.0f, 0.0f), L);
			Render();
			Fl::check();
		}
		return;

	}


	{

		PPC ppc1(*ppc);

		ppc->C = V3(20.0f, 100.0f, 200.0f);
		ppc->PositionAndOrient(ppc->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));

		for (int i = 0; i < 1000; i++) {
			ppc1.Pan(.5f);
			Render();
			ppc1.Visualize(ppc, fb, 20.0f);
			Fl::check();
		}
		return;
	}


	{

		for (int i = 0; i < 30; i++) {
			ppc->Pan(.5f);
			Render();
			Fl::check();
		}
		return;


	}


	{

		V3 tmc = tms[0].GetCenter();
		V3 C1 = ppc->C + V3(120.0f, 80.0f, 0.0f);
		ppc->PositionAndOrient(C1, tmc, V3(0.0f, 1.0f, 0.0));
		Render();
		return;

	}


	{
		TM tm;
		tm.LoadBin("geometry/teapot1K.bin");
		V3 tmC = ppc->C + ppc->GetVD()*100.0f;
		float tmSize = 50.0f;
		tm.PositionAndSize(tmC, tmSize);
		fb->SetBGR(0xFFFFFFFF);
		fb->ClearZB(0.0f);
		tm.RenderWireframe(ppc, fb);
		TM tm2;
//		tm2.SetRectangle(tmC, 50.0f, 50.0f);
//		tm2.RenderWireframe(ppc, fb);
		fb->redraw();

		for (int i = 0; i < 100; i++) {
			fb->SetBGR(0xFFFFFFFF);
			fb->ClearZB(0.0f);
			ppc->ChangeFocalLength(1.05f);
			tm.RenderWireframe(ppc, fb);
			fb->redraw();
			Fl::check();
		}

		return;
	}

	{

		V3 p(20.0f, -20.0f, -80.0f);
		V3 Oa(0.0f, -20.0f, -100.0f), a(0.0f, 1.0f, 0.0f);
		fb->Draw3DSegment(Oa, V3(0.0f, 0.0f, 0.0f), Oa + a * 30.0f, V3(0.0f, 1.0f, 0.0f), ppc);
		for (int i = 0; i < 270; i++) {
			p = p.RotateThisPointAboutArbitraryAxis(Oa, a, 1.0f);
			fb->Draw3DPoint(p, V3(1.0f, 0.0f, 0.0f), ppc, 7);
			fb->redraw();
			Fl::check();
		}
		return;

	}



	{
		TM tm;
		tm.SetRectangle(V3(0.0f, 0.0f, -100.0f), 45.0f, 30.0f);
//		tm.RenderPoints(ppc, fb);
		int stepsN = 360;
		for (int stepsi = 0; stepsi < stepsN; stepsi++) {
			fb->SetBGR(0xFFFFFFFF);
			tm.RotateAboutArbitraryAxis(V3(0.0f, 0.0f, -100.0f), V3(0.0f, 1.0f, 0.0f),
				1.0f);
			tm.RenderWireframe(ppc, fb);
			fb->redraw();
			Fl::check();
		}
		return;

	}

	{

		V3 p0(-20.0f, 0.0f, -100.0f), pp;
		V3 p1(-20.0f, 0.0f, -1000.0f);
		int stepsN = 5;
		for (int stepi = 0; stepi < stepsN; stepi++) {
//			fb->SetBGR(0xFFFFFFFF);
			V3 p = p0 + (p1 - p0) * ((float)stepi / (float)(stepsN - 1));
			fb->Draw3DPoint(p, V3(1.0f, 0.0f, 0.0f), ppc, 7);
			fb->redraw();
			Fl::check();
		}
		return;

	}


	{

		V3 p0(20.1f, 40.0f, 0.0f);
		V3 p1(50.9f, 240.0f, 0.0f);
		V3 c0(1.0f, 0.0f, 0.0f);
		V3 c1(0.0f, 1.0f, 0.0f);
		fb->DrawSegment(p0, c0, p1, c1);
		fb->redraw();
		return;
	}


	{
		V3 v(1.0f, 0.0f, 0.0f);
		for (int i = 0; i < 1000; i++) {
			float scf = (float)i / 999.0f;
			v[1] = scf;
			unsigned int color = v.GetColor();
			V3 v1; v1.SetFromColor(color);
			cerr << v1 - v << "          \r";
			fb->SetBGR(color);
			fb->redraw();
			Fl::check();
		}
		cerr << endl;
		return;
	}

	{
		int u0 = 100;
		int v0 = 50;
		int u1 = 300;
		int v1 = 400;
		int r = 30;
		int stepsN = 100;
		for (int stepi = 0; stepi < stepsN; stepi++) {
			fb->SetBGR(0xFFFFFFFF);
//			fb->DrawRectangle(u0 + stepi, v0, u1 + stepi, v1, 0xFF0000FF);
			fb->DrawCircle(u0 + stepi, v0, r, 0xFF0000FF);
			fb->redraw();
			Fl::check();
		}
		return;
	}

	{
		M33 m;
		m[0] = V3(1.0f, 0.0f, 0.0f);
		m[1] = V3(0.0f, 1.0f, 0.0f);
		m[2] = V3(0.0f, 0.0f, 1.0f);
		M33 m1;
		m1[0] = V3(-1.0f, 2.0f, 3.0f);
		m1[1] = V3(-2.0f, 2.0f, 3.0f);
		m1[2] = V3(-3.0f, -2.0f, 5.0f);

		cerr << m1 << endl << m * m1 << endl << m1 * m << endl;
		return;

		cerr << m << endl;
		V3 v0(1.0f, -2.0f, 0.5f);
		cerr << m * v0;
		cerr << v0;

		return;

	}

	{
		V3 v0(1.0f, -2.0f, 0.5f);
		V3 v1(-1.0f, -4.0f, 3.0f);
		cerr << v0 << v1;
		cerr << "v0v1 = " << v0 * v1 << endl;
		return;
	}

	V3 v(1.0f, 3.0f, -1.0f);
	cerr << v[1] << endl;
	v[1] = 2.0f;
	cerr << v[1] << endl;
	return;

	int u0 = 100;
	int v0 = 50;
	int u1 = 300;
	int v1 = 400;

	fb->DrawRectangle(u0, v0, u1, v1, 0xFF0000FF);
	fb->redraw();

	fb->SaveAsTiff("mydbg/rr.tif");
	
	FrameBuffer *fb1 = new FrameBuffer(30 + fb->w, 30, fb->w, fb->h);
	fb1->LoadTiff("mydbg/rr.tif");
	fb1->label("Loaded image");
	fb1->show();

	return;

	for (int u = 0; u < fb->w; u++) {
		fb->Set(u, fb->h / 2, 0xFF000000);
	}
	fb->redraw();
	cerr << "INFO: pressed DBG" << endl;

}